@@ -50,6 +50,12 @@ _Static_assert(sizeof(enum ei_keymap_type) == sizeof(int), "Invalid enum size");
 _Static_assert(sizeof(enum ei_event_type) == sizeof(int), "Invalid enum size");
 _Static_assert(sizeof(enum ei_log_priority) == sizeof(int), "Invalid enum size");
 
+static void
+ei_unsent_free(struct ei_unsent *unsent);
+
+static void
+ei_defunct_object_free(struct ei_defunct_object *obj);
+
 static void
 ei_destroy(struct ei *ei)
 {
@@ -59,14 +65,26 @@ ei_destroy(struct ei *ei)
 	while ((e = ei_get_event(ei)) != NULL)
 		ei_event_unref(e);
 
+	struct ei_unsent *unsent;
+	list_for_each_safe(unsent, &ei->unsent_queue, node) {
+		ei_unsent_free(unsent);
+	}
+
 	if (ei->backend_interface.destroy)
 		ei->backend_interface.destroy(ei, ei->backend);
+
 	ei->backend = NULL;
 	ei_handshake_unref(ei->handshake);
 	ei_connection_unref(ei->connection);
 	brei_context_unref(ei->brei);
+
 	sink_unref(ei->sink);
 	free(ei->name);
+
+	struct ei_defunct_object *obj;
+	list_for_each_safe(obj, &ei->defunct_objects, node) {
+		ei_defunct_object_free(obj);
+	}
 }
 
 static
@@ -100,6 +118,8 @@ ei_create_context(bool is_sender, void *user_data)
 	list_init(&ei->event_queue);
 	list_init(&ei->seats);
 	list_init(&ei->proto_objects);
+	list_init(&ei->unsent_queue);
+	list_init(&ei->defunct_objects);
 
 	ei->interface_versions = (struct ei_interface_versions){
 		.ei_connection = VERSION_V(1),
@@ -155,11 +175,23 @@ ei_register_object(struct ei *ei, struct brei_object *object)
 	list_append(&ei->proto_objects, &object->link);
 }
 
+static void
+ei_defunct_object_free(struct ei_defunct_object *obj)
+{
+	list_remove(&obj->node);
+	free(obj);
+}
+
 void
 ei_unregister_object(struct ei *ei, struct brei_object *object)
 {
 	log_debug(ei, "deregistering %s v%u object %#" PRIx64 "", object->interface->name, object->version, object->id);
 	list_remove(&object->link);
+
+	struct ei_defunct_object *obj = xalloc(sizeof *obj);
+	obj->object_id = object->id;
+	obj->time = ei_now(ei);
+	list_append(&ei->defunct_objects, &obj->node);
 }
 
 _public_ bool
@@ -657,6 +689,10 @@ handle_msg_disconnected(struct ei_connection *connection, uint32_t last_serial,
 
 	if (reason == EI_CONNECTION_DISCONNECT_REASON_DISCONNECTED) {
 		log_info(ei, "Disconnected by EIS");
+		/* We got disconnected, disconnect our source because whatever
+		 * we'd receive after this is garbage and the server won't
+		*  want to hear anything from us anyway. */
+		source_remove(ei->source);
 		ei_disconnect(ei);
 		return NULL;
 	} else {
@@ -667,11 +703,29 @@ handle_msg_disconnected(struct ei_connection *connection, uint32_t last_serial,
 }
 
 static struct brei_result *
-handle_msg_invalid_object(struct ei_connection *connection, uint32_t last_serial, object_id_t object)
+handle_msg_invalid_object(struct ei_connection *connection, uint32_t last_serial, object_id_t object_id)
 {
 	struct ei *ei = ei_connection_get_context(connection);
 
-	log_bug(ei, "Invalid object %#" PRIx64 " after %u, I don't yet know how to handle that", object, last_serial);
+	/* The protocol is async, so what can happen is:
+	 *
+	 * server sends A->destroyed()
+	 *     client sends A->foo()
+	 *     client receives A->destroyed()
+	 * server receives A->foo()
+	 * server sends invalid_object_id(A)
+	 *     client receives invalid_object_id(A)
+	 *
+	 * This is expected and we shouldn't have a problem with that.
+	 */
+
+	struct ei_defunct_object *defunct;
+	list_for_each_safe(defunct, &ei->defunct_objects, node) {
+		if (defunct->object_id == object_id)
+			return NULL;
+	}
+
+	log_bug(ei, "Invalid object %#" PRIx64 " after serial %u, I don't yet know how to handle that", object_id, last_serial);
 
 	return NULL;
 }
@@ -726,15 +780,33 @@ lookup_object(object_id_t object_id, struct brei_object **object, void *userdata
 static void
 connection_dispatch(struct source *source, void *userdata)
 {
+	static uint8_t cleanup;
 	struct ei *ei = userdata;
 	enum ei_state old_state = ei->state;
 
-	_unref_(brei_result) *result = brei_dispatch(ei->brei, source_get_fd(source),
-						     lookup_object, ei);
-	if (result) {
-		log_warn(ei, "Connection error: %s", brei_result_get_explanation(result));
-		brei_drain_fd(source_get_fd(source));
+	/* Flush any pending writes, if we have them */
+	int rc = ei_unsent_flush(ei);
+	if (rc < 0 && rc != -EAGAIN) {
+		log_warn(ei, "Error flushing unsent queue: %s", strerror(-rc));
 		ei_disconnect(ei);
+	} else {
+		_unref_(brei_result) *result = brei_dispatch(ei->brei, source_get_fd(source),
+							     lookup_object, ei);
+		if (result) {
+			log_warn(ei, "Connection error: %s", brei_result_get_explanation(result));
+			brei_drain_fd(source_get_fd(source));
+			ei_disconnect(ei);
+		} else if (++cleanup % 20 == 0) {
+			uint64_t now = ei_now(ei);
+			struct ei_defunct_object *defunct;
+
+			list_for_each_safe(defunct, &ei->defunct_objects, node) {
+				/* Drop defunct objects after 5s */
+				if (now - defunct->time < s2us(5))
+					break;
+				ei_defunct_object_free(defunct);
+			}
+		}
 	}
 
 	static const char *states[] = {
@@ -752,6 +824,45 @@ connection_dispatch(struct source *source, void *userdata)
 			  states[ei->state]);
 }
 
+static void
+ei_queue_unsent(struct ei *ei, struct source *source, struct iobuf *buf)
+{
+	if (list_empty(&ei->unsent_queue)) {
+		source_enable_write(source, true);
+	}
+
+	struct ei_unsent *unsent = xalloc(sizeof *unsent);
+	unsent->buf = buf;
+	list_append(&ei->unsent_queue, &unsent->node);
+}
+
+static void
+ei_unsent_free(struct ei_unsent *unsent)
+{
+	list_remove(&unsent->node);
+	iobuf_free(unsent->buf);
+	free(unsent);
+}
+
+int
+ei_unsent_flush(struct ei* ei)
+{
+	if (list_empty(&ei->unsent_queue))
+		return 0;
+
+	struct source *source = ei->source;
+	int fd = source_get_fd(source);
+	struct ei_unsent *unsent;
+	list_for_each_safe(unsent, &ei->unsent_queue, node) {
+		int rc = iobuf_send(unsent->buf, fd);
+		if (rc < 0)
+			return rc;
+		ei_unsent_free(unsent);
+	}
+	source_enable_write(source, false);
+	return 0;
+}
+
 int
 ei_send_message(struct ei *ei, const struct brei_object *object,
 		uint32_t opcode, const char *signature, size_t nargs, ...)
@@ -781,7 +892,19 @@ ei_send_message(struct ei *ei, const struct brei_object *object,
 	_cleanup_iobuf_ struct iobuf *buf = brei_result_get_data(result);
 	assert(buf);
 	int fd = source_get_fd(ei->source);
-	int rc = iobuf_send(buf, fd);
+	int rc = -EPIPE;
+	if (fd != -1) {
+		rc = ei_unsent_flush(ei);
+		if (rc >= 0)
+			rc = iobuf_send(buf, fd);
+		if (rc == -EAGAIN) {
+			ei_queue_unsent(ei, ei->source, steal(&buf));
+			rc = 0;
+		} else if (rc < 0){
+			log_warn(ei, "failed to send message: %s", strerror(-rc));
+			source_remove(ei->source);
+		}
+	}
 	return rc < 0 ? rc : 0;
 }
 
@@ -825,19 +948,30 @@ ei_configure_name(struct ei *ei, const char *name)
 	ei->name = xstrdup(name);
 }
 
+_public_ void
+ei_clock_set_now_func(struct ei *ei, ei_clock_now_func func)
+{
+	ei->clock_now = func;
+}
+
 _public_ uint64_t
 ei_now(struct ei *ei)
 {
 	uint64_t ts = 0;
-	int rc = now(&ts);
 
-	if (rc < 0) {
-		/* We should probably disconnect here but the chances of this
-		 * happening are so slim it's not worth worrying about. Plus,
-		 * if this fails we're likely to be inside eis_device_frame()
-		 * so we should flush a frame event before disconnecting and... */
-		log_error(ei, "clock_gettime failed: %s", strerror(-rc));
+	if (ei->clock_now)
+		ts = ei->clock_now(ei);
+	else {
+		int rc = now(&ts);
+		if (rc < 0) {
+			/* We should probably disconnect here but the chances of this
+			 * happening are so slim it's not worth worrying about. Plus,
+			 * if this fails we're likely to be inside eis_device_frame()
+			 * so we should flush a frame event before disconnecting and... */
+			log_error(ei, "clock_gettime failed: %s", strerror(-rc));
+		}
 	}
+
 	return ts;
 }
 
