@@ -216,6 +216,9 @@ eis_device_in_region(struct eis_device *device, double x, double y)
 {
 	struct eis_region *r;
 
+	if (list_empty(&device->regions))
+		return true;
+
 	list_for_each(r, &device->regions, link) {
 		if (eis_region_contains(r, x, y))
 			return true;
@@ -1041,12 +1044,8 @@ eis_device_pointer_motion_absolute(struct eis_device *device,
 	if (device->state != EIS_DEVICE_STATE_EMULATING)
 		return;
 
-	struct eis_region *r;
-	list_for_each(r, &device->regions, link) {
-		if (!eis_region_contains(r, x, y)) {
-			return;
-		}
-	}
+	if (!eis_device_in_region(device, x, y))
+		return;
 
 	device->send_frame_event = true;
 
@@ -1263,14 +1262,11 @@ eis_touch_down(struct eis_touch *touch, double x, double y)
 		return;
 	}
 
-	struct eis_region *r;
-	list_for_each(r, &device->regions, link) {
-		if (!eis_region_contains(r, x, y)) {
-			log_bug_client(eis_device_get_context(device),
-				       "%s: touch %u has invalid x/y coordinates", __func__, touch->tracking_id);
-			touch->state = TOUCH_IS_UP;
-			return;
-		}
+	if (!eis_device_in_region(device, x, y)) {
+		log_bug_client(eis_device_get_context(device),
+			       "%s: touch %u has invalid x/y coordinates", __func__, touch->tracking_id);
+		touch->state = TOUCH_IS_UP;
+		return;
 	}
 
 	touch->state = TOUCH_IS_DOWN;
@@ -1286,14 +1282,11 @@ eis_touch_motion(struct eis_touch *touch, double x, double y)
 		return;
 
 	struct eis_device *device = eis_touch_get_device(touch);
-	struct eis_region *r;
-	list_for_each(r, &device->regions, link) {
-		if (!eis_region_contains(r, x, y)) {
-			log_bug_client(eis_device_get_context(device),
-				       "%s: touch %u has invalid x/y coordinates", __func__, touch->tracking_id);
-			eis_touch_up(touch);
-			return;
-		}
+	if (!eis_device_in_region(device, x, y)) {
+		log_bug_client(eis_device_get_context(device),
+			       "%s: touch %u has invalid x/y coordinates", __func__, touch->tracking_id);
+		eis_touch_up(touch);
+		return;
 	}
 
 	device->send_frame_event = true;
